package de.fraunhofer.iem.dataprovider.toolRun

import de.fraunhofer.iem.dataprovider.taskManager.tasks.GitRepository
import de.fraunhofer.iem.dataprovider.taskManager.tasks.toDbObject
import de.fraunhofer.iem.dataprovider.toolRun.model.*
import org.springframework.stereotype.Service
import java.util.*

@Service
class ToolRunService(
    private val repositoryRepository: RepositoryRepository,
    private val toolRunRepository: ToolRunRepository,
    private val toolRepository: ToolRepository,
    private val ruleRepository: RuleRepository
) {

    fun getLatestToolRunsForRepo(repoId: UUID) {
        repositoryRepository.findById(repoId).ifPresent { repo ->
            val latestToolRuns = repo.tools.map { tool ->
                toolRunRepository.findFirstByToolIdAndRepositoryIdOrderByTimeStampDesc(tool.id!!, repoId)
            }
            latestToolRuns.forEach { tr ->
                println("TR Time: ${tr.timeStamp}")
            }
            // TODO: here we can calculate the metrics
        }
    }

    // TODO: this func should be a suspend fun. We can achieve this by
    //  wrapping the db return types in flux, mono, or futures
    fun saveSarif(sarif: Sarif, repoId: UUID) {

        val repo = repositoryRepository.findById(repoId)

        sarif.runs.forEach { run ->

            val tool = findOrCreateTool(run.tool)
            val rules = findRulesForTool(tool.id)

            createNewRulesAndSaveToTool(run.tool.driver.rules, rules, tool)


            val toolResults = run.results.map { sarifResult ->
                val tr = sarifResult.asDbObject()
                // link rules to results
                rules.filter { it.sarifRuleId == sarifResult.ruleId }
                    .forEach { it.addResultLocation(tr) }
                tr
            }


            val toolRun = run.asDbObject()
            toolRun.tool = tool
            toolRun.addRules(rules)
            toolRun.addFindings(toolResults)

            val result = toolRunRepository.save(toolRun)
            toolRepository.save(tool)

            repo.ifPresent {
                if (!it.tools.contains(tool)) {
                    it.tools.add(tool)
                }
                it.addToolResult(result)
                repositoryRepository.save(it)
            }
        }
    }

    private fun createNewRulesAndSaveToTool(
        sarifRules: List<SarifRule>,
        rules: MutableList<Rule>,
        tool: Tool
    ) {
        sarifRules.forEach { sarifRule ->
            if (rules.find { it.sarifRuleId == sarifRule.id } == null) {
                val rule = ruleRepository.save(sarifRule.asDbObject())
                rules.add(rule)
                tool.addRule(rule)
            }
        }
    }

    private fun findOrCreateTool(tool: SarifTool): Tool {
        val fullName = tool.driver.fullName
        val name = tool.driver.name
        val version = tool.driver.version

        return toolRepository.findByFullNameIgnoreCaseAndNameIgnoreCaseAndVersionIgnoreCase(fullName, name, version)
            ?: toolRepository.save(tool.asDbObject())
    }

    private fun findRulesForTool(id: UUID?): MutableList<Rule> {
        return if (id != null) {
            ruleRepository.findByTool_Id(id).toMutableList()
        } else {
            mutableListOf()
        }
    }

    fun getRepoOrCreate(gitRepository: GitRepository): Repository {
        var repo = repositoryRepository.findByRepoIdAndPlatform(gitRepository.id, gitRepository.platform)
        if (repo == null) {
            repo = repositoryRepository.save(gitRepository.toDbObject())
        }
        return repo
    }
}
